var cmp = require('../util/compare.js');
var clazz = require('../../core/clazz.js');
var type = require('../../core/type.js');
var Observable = require('../Observable.js');

var RED = 'RED', BLACK = 'BLACK', NIL = {
    color: BLACK,
    isNil: function() {
        return true;
    },
    setLeft: function(left) {
        this.left = left;
    },
    setRight: function(right) {
        this.right = right;
    },
    setParent: function(parent) {
        this.parent = parent;
    }
};

function max(node) {
    while(!node.right.isNil()) {
        node = node.right;
    }
    return node;
}

function min(node) {
    while(!node.left.isNil()) {
        node = node.left;
    }
    return node;
}

function successor(node) {
    var sucsor = node.parent;
    if (!node.right.isNil()) {
        return min(node.right);
    }
    while(!sucsor.isNil() && node === sucsor.right) {
        node = sucsor;
        sucsor = sucsor.parent;
    }
    return sucsor;
}

function predecessor(node) {
    var pred = node.parent;
    if (!node.left.isNil()) {
        return max(node.left);
    }
    while(!pred.isNil() && node === pred.left) {
        node = pred;
        pred = pred.parent;
    }
    return pred;
}

function transplant(tree, node, replacement) {
    if (node.parent.isNil()) {
        tree.root = replacement;
    } else if (node === node.parent.left) {
        node.parent.setLeft(replacement);
    } else {
        node.parent.setRight(replacement);
    }
    replacement.setParent(node.parent);
}

function leftRotate(tree, node) {
    var y = node.right;
    node.setRight(y.left);
    if (!y.left.isNil()) {
        y.left.setParent(node);
    }
    y.setParent(node.parent);
    if (node.parent.isNil()) {
        tree.root = y;
    } else if (node === node.parent.left) {
        node.parent.setLeft(y);
    } else {
        node.parent.setRight(y);
    }
    y.setLeft(node);
    node.setParent(y);
}

function rightRotate(tree, node) {
    var y = node.left;
    node.setLeft(y.right);
    if (!y.right.isNil()) {
        y.right.setParent(node);
    }
    y.setParent(node.parent);
    if (node.parent.isNil()) {
        tree.root = y;
    } else if (node === node.parent.right) {
        node.parent.setRight(y);
    } else {
        node.parent.setLeft(y);
    }
    y.setRight(node);
    node.setParent(y);
}

function map(node, filter, scope) {
    var elements = [];
    if (!node.isNil()) {
        elements = elements.concat(map(node.left, filter, scope));
        if (filter.call(scope, node.value) !== false) {
            elements.push(node.value);
        }
        elements = elements.concat(map(node.right, filter, scope));
    }
    return elements;
}

function insertFixup(tree, node) {
    var y;
    while(node.parent.color === RED) {
        if (node.parent === node.parent.parent.left) {
            y = node.parent.parent.right;
            if (y.color === RED) {
                node.parent.color = BLACK;
                y.color = BLACK;
                node.parent.parent.color = RED;
                node = node.parent.parent;
            } else if (node === node.parent.right) {
                node = node.parent;
                leftRotate(tree, node);
            } else {
                node.parent.color = BLACK;
                node.parent.parent.color = RED;
                rightRotate(tree, node.parent.parent);
            }
        } else {
            y = node.parent.parent.left;
            if (y.color === RED) {
                node.parent.color = BLACK;
                y.color = BLACK;
                node.parent.parent.color = RED;
                node = node.parent.parent;
            } else if (node === node.parent.left) {
                node = node.parent;
                rightRotate(tree, node);
            } else {
                node.parent.color = BLACK;
                node.parent.parent.color = RED;
                leftRotate(tree, node.parent.parent);
            }
        }
    }
    tree.root.color = BLACK;
}

function deleteFixup(tree, node) {
    var w;
    while(node !== tree.root && node.color === BLACK) {
        if (node === node.parent.left) {
            w = node.parent.right;
            if (w.color === RED) {
                w.color = BLACK;
                node.parent.color = RED;
                leftRotate(tree, node.parent);
                w = node.parent.right;
            }
            if (w.left.color === BLACK && w.right.color === BLACK) {
                w.color = RED;
                node = node.parent;
            } else if (w.right.color === BLACK) {
                w.left.color = BLACK;
                w.color = RED;
                rightRotate(tree, w);
                w = node.parent.right;
            } else {
                w.color = node.parent.color;
                node.parent.color = BLACK;
                w.right.color = BLACK;
                leftRotate(tree, node.parent);
                node = tree.root;
            }
        } else {
            w = node.parent.left;
            if (w.color === RED) {
                w.color = BLACK;
                node.parent.color = RED;
                rightRotate(tree, node.parent);
                w = node.parent.left;
            }
            if (w.right.color === BLACK && w.left.color === BLACK) {
                w.color = RED;
                node = node.parent;
            } else if (w.left.color === BLACK) {
                w.right.color = BLACK;
                w.color = RED;
                leftRotate(tree, w);
                w = node.parent.left;
            } else {
                w.color = node.parent.color;
                node.parent.color = BLACK;
                w.left.color = BLACK;
                rightRotate(tree, node.parent);
                node = tree.root;
            }
        }
    }
    node.color = BLACK;
}

var RedBlackTree = clazz.define(Observable, {
    compare: cmp,
    root: NIL,
    size: 0,
    constructor: function RedBlackTree(compare) {
        if (type.isFunction(compare)) {
            this.compare = compare;
        }
    },
    newNode: function(value) {
        return new Node(value, this);
    },
    insert: function(value) {
        var y = NIL, x = this.root, node = this.newNode(value);
        while(!x.isNil()) {
            y = x;
            if (this.compare(node.value, x.value) < 0) {
                x = x.left;
            } else {
                x = x.right;
            }
        }
        node.setParent(y);
        if (y.isNil()) {
            this.root = node;
        } else if (this.compare(node.value, y.value) < 0) {
            y.setLeft(node);
        } else {
            y.setRight(node);
        }
        this.emit('nodeWillAttach', node);
        insertFixup(this, node);
        this.emit('nodeDidAttach', node);
        this.size++;
        return this;
    },
    delete: function(value) {
        var node = this.get(value);
        var y = node, original = y.color;
        var x;
        if (node.isNil()) {
            return;
        }
        this.emit('nodeWillDetach', node);
        if (node.left.isNil()) {
            x = node.right;
            transplant(this, node, node.right);
        } else if (node.right.isNil()) {
            x = node.left;
            transplant(this, node, node.left);
        } else {
            y = min(node.right);
            original = y.color;
            x = y.right;
            if (y.parent === node) {
                x.setParent(y);
            } else {
                transplant(this, y, y.right);
                y.setRight(node.right);
                y.right.setParent(y);
            }
            transplant(this, node, y);
            y.setLeft(node.left);
            y.left.setParent(y);
            y.color = node.color;
        }
        if (original === BLACK) {
            deleteFixup(this, x);
        }
        this.emit('nodeDidDetach', node);
        this.size--;
        return node.value;
    },
    get: function(value) {
        var node = this.root, cmp = 0;
        while(!node.isNil() && (cmp = this.compare(node.value, value)) !== 0) {
            node = cmp > 0 ? node.left : node.right;
        }
        return node;
    },
    map: function(filter, scope) {
        return map(this.root, type.isFunction(filter) ? filter : clazz.emptyFn, scope);
    },
    each: function(callback, scope) {
        var p = this.min();
        while(!p.isNil() && callback.call(scope, p.value) !== false) {
            p = p.successor();
        }
    },
    min: function() {
        return min(this.root);
    },
    max: function() {
        return max(this.root);
    }
});

var Node = clazz.define(Observable, {
    color: RED,
    left: NIL,
    right: NIL,
    parent: NIL,
    value: {},
    tree: {},
    constructor: function Node(value, tree) {
        this.value = value;
        this.tree = tree;
    },
    successor: function() {
        return this.isNil() ? this : successor(this);
    },
    predecessor: function() {
        return this.isNil() ? this : predecessor(this);
    },
    isNil: function() {
        return false;
    },
    isLeaf: function() {
        return this.left.isNil() && this.right.isNil();
    },
    sibling: function() {
        if (this.parent.isNil()) {
            return NIL;
        }
        if (this === this.parent.left) {
            return this.parent.right;
        } else {
            return this.parent.left;
        }
    },
    setLeft: function(left) {
        this.emit('childWillChange', left);
        this.left = left;
        this.emit('childDidChange');
    },
    setRight: function(right) {
        this.emit('childWillChange', right);
        this.right = right;
        this.emit('childDidChange');
    },
    setParent: function(parent) {
        this.parent = parent;
    }
});

RedBlackTree.Node = Node;

module.exports = RedBlackTree;